//============================================================================
// Name        : binauto.c
// Author      : 
// Version     :
// Copyright   : Copyright (c) 2014 by Mehdi Alizadeh
// Description :
//============================================================================

#include "BA.h"

int main(int argc, char* argv[])
{
	// BEGIN DEBUG
	// uncomment these lines to test different layers
	// Ftest(); return 0;
	// Ztest(); return 0;
	// fastKNN(10000000, 64); return 0;
	// END DEBUG

	MPI_Init(&argc, &argv); 					// initialize the MPI environment
	int mpirank;								// the rank of the current process
	MPI_Comm_rank(MPI_COMM_WORLD, &mpirank);	// get mpirank
	int mpisize;								// total number of MPI processes
	MPI_Comm_size(MPI_COMM_WORLD, &mpisize);	// get mpisize

	// path to the configurations file is read from application argument input
	configstruct *config = NULL;
	if (argc > 1)
		config = loadconfig(argv[1], mpirank == ROOT_PROCESS); // read and set config struct

	if (!config) // if something went wrong in reading the config file terminate the program
	{
		if (mpirank == ROOT_PROCESS)
			printf("missing or invalid argument. binautompi will exit now.\n");
		MPI_Finalize();
		return EXIT_SUCCESS;
	}

	// BEGIN DEBUG
	// translate the reporterrs flag
	int w = config->reporterrs;
	int report_inter_error = (w == 2 || w == 3) ? 1 : 0; 			// report intermediate errors
	int report_final_error = (w == 1 || w == 2 || w == 3) ? 1 : 0;	// report the final errors
	int report_inter_precision = (w == 2) ? 1 : 0;					// report intermediate precisions
	int report_final_precision = (w == 2 || w == 3) ? 1 : 0;		// report the final precision
	int calculate_ranks = (w == 4 || w == 6) ? 1 : 0;	// calculate ranks based on previously trained models
	int calculate_AEerr = (w == 5 || w == 6) ? 1 : 0;	// calculate nested errors based on previously trained models
	// END DEBUG

	int maxit = config->maxit; 							// total iterations, size of mu array
	double* muvalues = config->mu;						// mu values
	long int totalminibatches = config->minibatchcnt;	// total number of minibatches
	long int minibatches = totalminibatches / mpisize;	// minibatches per processor
	int ringepochs = config->epochs;					// number of epochs in each W-step
	int localepochs = 1;								// we can do extra epochs over the local data, but to
														// keep things simple we don't use local epochs (set to 1)

	// load shuffling lookup matrix based on the current settings
	// if we don't find a matching shuffling file the default permutation (no shuffling) will be loaded
	int* permutation = loadperm(NULL/*"shuffle/shuffle-26-8-32-384"*/, maxit, ringepochs, mpisize, totalminibatches, mpirank == ROOT_PROCESS);

	// allocate buffers to store different metrics over iterations
	double Herr[maxit];				// H layer error over iterations on local data
	double Ferr[maxit];				// F layer error over iterations on local data
	double Nestederr[maxit];		// Nested error of the model over iterations on local data
	double TOTALobjecterr[maxit];	// Aggregated objective error from all the processes
	double TOTALnestederr[maxit];	// Aggregated nested error from all the processes
	double TOTALprecision[maxit + 1];	// Total precision combined from all the processes
	double itertime[maxit];		// iteration duration
	double commtime[maxit];		// communication time in each iteration
	double wstptime[maxit];		// W-step time in each iteration (f + h)
	double zstptime[maxit];		// Z-step time in each iteration
	double wstptime_h[maxit];	// h-train time in each iteration
	double wstptime_f[maxit];	// f-train time in each iteration
	int i;
	for (i = 0; i < maxit; i++)
		TOTALobjecterr[i] = TOTALnestederr[i] = TOTALprecision[i] = -1;
	TOTALprecision[maxit] = -1;

	intype *Xin = NULL; 	// input matrix (NxDin)
	outype *Xout = NULL;	// output matrix (NxDout)
	biguint *Z = NULL;		// auxiliary variables (N biguints)
	long int N = loaddatasets(config, &Xin, &Xout, &Z, mpirank, mpisize, 1); // load datasets

	int Din = config->Din;		// input dimension
	int Dout = config->Dout;	// output dimension
	int L = config->L;			// code size, number of bits

	// if the training time for the f layer is large, we can achieve better parallel performance
	// by breaking output nodes in the decoder layer to individual submodels.
	int breakf = 1; 	// 0: consider the whole f layer as one unit, 1: separate each output node in f layer
	int useBsend = 1;	// 0: use asynchronous send and wait before the next send, 1: use MPI_Bsend (buffered send)

	// Initialization layers and allocate memory
	HinitSGD(Din, L); 		// initialize encoder
	Finit(L, Dout, breakf);	// initialize decoder
	Zinit(1, Dout, L);		// initialize variables in Z-step

	// allocate memory for output of the hash functions
	biguint* hX = (biguint*)calloc(N, sizeof(biguint));

	int totaltaskcount = L + (breakf ? Dout : 1);	// calculate the number of submodels
	int taskgroups = 0;								// we can group submodels in taskgroups for performance gains
	int **tasks = dividetasks(totaltaskcount, mpisize, &taskgroups); // arrange submodels in task groups
	int *buffersizes = calcbuffersizes(Din, Dout, L, taskgroups, tasks, breakf); // calc buffer size for each task group
	int maxbufsize = 0;	// maximum buffer size of all taskgroups
	int modelsize = 0; 	// total model size
	for (i = 0; i < taskgroups; i++)
	{
		if (buffersizes[i] > maxbufsize)	// find the maximum buffer size
			maxbufsize = buffersizes[i];
		modelsize += buffersizes[i];		// sum up the total model size
	}

	// BEGIN DEBUG
	// allocate buffer to check the count of visits for each submodel, debug purpose
	int *taskvisits = calloc(totaltaskcount, sizeof(int));
	// END DEBUG

	// calculate the communication buffer size
	// in each buffer the first elements is the submodel index and the second element is the stepcounter
	int commbuffsize = maxbufsize + 2 * sizeof(double);

	// allocate buffers for send and receive
	double *sendbuffer = (double*) malloc(commbuffsize);	// send buffer
	double *receivebuffer = (double*) malloc(commbuffsize);	// receive buffer
	double *modelbuffer = (double*) malloc(modelsize);		// the buffer holding the total model

	if (calculate_ranks || calculate_AEerr) // if we are only interested in processing previously trained models
	{
		// just load previously trained models from files and process them, print the results and exit.
		return process_trained_models(calculate_ranks, calculate_AEerr, config,
				mpirank, modelbuffer, modelsize, totaltaskcount, Xin, Xout, Z, N, Nestederr, TOTALnestederr);
	}

	// request and status structs to use while waiting for asynchronous send and receive
	MPI_Request sendRequest;	// used in case of asynchronous send
	MPI_Request recvRequest;	// used for asynchronous receive
	MPI_Status recvStatus;		// used to check the size of data received in a receive command

	int mpi_attach_buff_size = 0;	// size of buffer to attach to MPI environment for buffered send
	void *mpi_attach_buff = NULL;	// the buffer we attach to MPI environment used in buffered send
	if (useBsend) // if we are using buffered send
	{
		// we use MPI_Bsend to avoid buffer management complications, so we need to allocation buffer for it
		// calculate the buffer size needed
		MPI_Pack_size(commbuffsize, MPI_CHAR, MPI_COMM_WORLD, &mpi_attach_buff_size);
		// adding header size and making sure that the whole model will fit in the attached buffer
		mpi_attach_buff_size = totaltaskcount * (mpi_attach_buff_size + MPI_BSEND_OVERHEAD);
		 // allocating the memory for attach buffer
		mpi_attach_buff = malloc(mpi_attach_buff_size);
		 // attaching the buffer to MPI environment
		MPI_Buffer_attach(mpi_attach_buff, mpi_attach_buff_size);
	}

	// variables to store current times, durations will be calculated as (endtime - begintime)
	struct timeval iterbegin, iterend;
	struct timeval commbegin, commend;
	struct timeval wstpbegin, wstpend;
	struct timeval zstpbegin, zstpend;
	struct timeval wstpbegin_, wstpend_;

	// BEGIN DEBUG
	// print the status of processes and loaded data sizes. Also some info about the initial status
	if (!synchronizeprocesses() && mpirank == ROOT_PROCESS)
	{
		printf("all processes initialized successfully.\n");
		printf("bitwise calculations are available for %dbit variables.\n", (int)(sizeof(biguint) * 8));
		int n = N / minibatches;
		printf("total %ld minibatches (%ld per processor, %ld requested) of size %d~%d.\n",
				minibatches * mpisize, minibatches, totalminibatches, n, n + 1);

		// print the arranges taskgroups and submodels in each of them
		int igourp, itask;
		for (igourp = 0; igourp < taskgroups; igourp++)
		{
			printf("S%02d tasks: [ ", igourp);
			itask = 0;
			while (tasks[igourp][itask] != -1)
			{
				if (tasks[igourp][itask] < L)
					printf("SVM%02d ", (int) (tasks[igourp][itask]));
				else
					printf("ANN%02d ", (int) (tasks[igourp][itask]) - L);

				itask++;
			}
			printf("] buf-size: %d bytes.\n", buffersizes[igourp]);
		}

		if (report_inter_precision) // calculate initial precision
			TOTALprecision[0] = calcprecision(config, 1);
	}
	// END DEBUG

	int iter;
	for (iter = 0; iter < maxit; iter++) // the main loop over iterations
	{
		// initializing time keeper variables for the current iteration
		commtime[iter] = wstptime[iter] = 0;
		wstptime_h[iter] = wstptime_f[iter] = 0;

		// uncomment the following line if you want to synch all the processes in the beginning if each iteration
		// synchronizeprocesses();

		gettimeofday(&iterbegin, NULL);

		double mu = muvalues[iter]; 	// mu value used in this iteration
		if (mpirank == ROOT_PROCESS) printf("iteration %02d mu = %12.8f\n", iter + 1, mu);

		// initializing the variables before W-step
		int tasksindex = -1, *steptasks = NULL;
		int taskbufsize = 0, stepcounter = 0;

		int counter = 0; // counts the number of submodels this process has visited in this iteration
		// W-STEP BEGINS
		while (counter < (ringepochs + 1) * taskgroups) // stop if we have processed all the submodels
		{
			if (counter == 0) // if this is the first loop in W-step
			{
				// select the task, is the number of tasks are less than number of processes
				// we may not need to do anything until we receive a task from previous process
				tasksindex = mpirank < taskgroups ? mpirank : -1;
				steptasks = tasksindex >= 0 ? tasks[tasksindex] : NULL; // select the current task list
				stepcounter = 0;
				taskbufsize = tasksindex >= 0 ? buffersizes[tasksindex] : 0; // select the current buffer size
			}

			// RECEIVE MODEL BEGINS
			if (counter > 0 || tasksindex < 0) // if should wait to receive a submodel
			{
				gettimeofday(&commbegin, NULL);

				// activate an asynchronous receive, accept messages from any possible source.
				MPI_Irecv(receivebuffer, commbuffsize, MPI_CHAR, MPI_ANY_SOURCE, MODEL_MSG_TAG, MPI_COMM_WORLD, &recvRequest);
				MPI_Wait(&recvRequest, &recvStatus); // wait until we actually receive something

				tasksindex = (int)receivebuffer[0]; 		// the first element in the buffer is the task index
				steptasks = tasks[tasksindex];				// choose tasks to process
				stepcounter = (int)receivebuffer[1];		// the second element in the buffer is stepcounter
				double *tasksbuffer = &(receivebuffer[2]);	// separate the actual model buffer
				taskbufsize = buffersizes[tasksindex];		// the size we expect to receive

				// DEBUG BEGIN
				// check if the size we have received is the size we expected, if not print a warning message
				int receivedcount = 0;
				MPI_Get_count(&recvStatus, MPI_CHAR, &receivedcount);
				if (receivedcount != taskbufsize + 2 * sizeof(double)) // buffer size is submodel size + 2
					printf("received count %d expected count %d\n", receivedcount, (int)(taskbufsize + 2 * sizeof(double)));
				// DEBUG END

				int index = 0;
				int itask = 0;
				while (steptasks[itask] != -1)	// save all the submodels from the received buffer into
				{								// their adequate structures to be able to train later
					
					if (steptasks[itask] < L)
						setHmodel(steptasks[itask], tasksbuffer, &index);		// this is a SVM
					else
						setFmodel(steptasks[itask] - L, tasksbuffer, &index);   // this is a linear layer
					itask++;
				}

				if (index * sizeof(double) != taskbufsize) // DEBUG ONLY: if something went wrong in saving the buffer
					printf("inconsistent setmodel %d %d\n", index * (int)sizeof(double), taskbufsize);

				gettimeofday(&commend, NULL);
				commtime[iter] += HighResDiff(commend, commbegin);
			} // RECEIVE MODEL ENDS

			// TRAIN MODEL BEGINS
			if (stepcounter < ringepochs * mpisize) // if it is not the last loop in W-step
			{										// in the last loop we only pass the submodels without training
				int iepoch = stepcounter / mpisize; // calc the current epoch for this submodel
				// load the minibatch shuffling lookup table for this submodel
				int *minibatchperm = &(permutation[(iter * ringepochs + iepoch) * (mpisize + minibatches) + mpisize]);

				gettimeofday(&wstpbegin, NULL);

				int itask = 0;
				while (steptasks[itask] != -1) // train all the submodels in this taskgroup
				{
					taskvisits[steptasks[itask]]++; // DEBUG ONLY: count the number of visits
					gettimeofday(&wstpbegin_, NULL);

					if (steptasks[itask] < L)	// this is a SVM
						HtrainSGD(Xin, Z, N, Din, minibatches, localepochs, steptasks[itask], stepcounter, minibatchperm);
					else						// this is a linear layer
						FtrainSGD(Z, Xout, N, Dout, L, minibatches, localepochs, steptasks[itask] - L, stepcounter, minibatchperm);

					gettimeofday(&wstpend_, NULL);
					if (steptasks[itask] < L)
						wstptime_h[iter] += HighResDiff(wstpend_, wstpbegin_);
					else
						wstptime_f[iter] += HighResDiff(wstpend_, wstpbegin_);

					itask++;
				}

				gettimeofday(&wstpend, NULL);
				wstptime[iter] += HighResDiff(wstpend, wstpbegin);
			} // TRAIN MODEL ENDS

			// SEND MODEL BEGINS
			if (stepcounter < (ringepochs + 1) * mpisize - 1)	// based on the stepcount of the submodel we
			{													// decide to stop or send to the next process
				gettimeofday(&commbegin, NULL);

				// in case of asynchronous send we should wait BEFORE modifying the buffer
				if (!useBsend && counter > 0)
					MPI_Wait(&sendRequest, MPI_STATUS_IGNORE);

				double *tasksbuffer = &(sendbuffer[2]);

				//*
				// send model to the next process
				int index = 0;
				int itask = 0;
				while (steptasks[itask] != -1) // load submodels from their adequate structures into send buffer
				{
					if (steptasks[itask] < L)
						getHmodel(steptasks[itask], tasksbuffer, &index);		// this is a SVM
					else
						getFmodel(steptasks[itask] - L, tasksbuffer, &index);	// this is a linear layer

					itask++;
				}

				// DEBUG BEGIN
				if (index * sizeof(double) != taskbufsize) // check if something went wrong in load data
					printf("inconsistent getmodel %d %d\n", index * (int)sizeof(double), taskbufsize);
				// DEBUG END

				int iepoch = (stepcounter + 1) / mpisize; // find the next epoch index for this submodel
				if (iepoch == ringepochs) 	// the last loop is only for propagate the final trained model
					iepoch--;				// so we use the shuffling from previous round

				// load the shuffling lookup table for this submodel in current epoch
				int *rankperm = &(permutation[(iter * ringepochs + iepoch) * (mpisize + minibatches)]);
				int posinring = -1;
				for (i = 0; i < mpisize; i++) // search in the table to find current process
					if (mpirank == rankperm[i])
					{
						posinring = i;
						break;
					}
				// choose next process to send the submodel to
				int nextrank = posinring == mpisize - 1 ? rankperm[0] : rankperm[posinring + 1];

				sendbuffer[0] = tasksindex; // set the task index in the first element of the buffer
				stepcounter++;				// increase the stepcounter for the current submodel
				sendbuffer[1] = stepcounter;// set the stepcounter in the second element of the buffer

				if (useBsend)	// send using buffered MPI send
					MPI_Bsend(sendbuffer, taskbufsize + 2 * sizeof(double), MPI_CHAR, nextrank, MODEL_MSG_TAG, MPI_COMM_WORLD);
				else			// send using synchronous MPI send
					MPI_Isend(sendbuffer, taskbufsize + 2 * sizeof(double), MPI_CHAR, nextrank, MODEL_MSG_TAG, MPI_COMM_WORLD, &sendRequest);

				gettimeofday(&commend, NULL);
				commtime[iter] += HighResDiff(commend, commbegin);
			} // SEND MODEL ENDS
			counter++;
		} // W-STEP ENDS

		if (!useBsend) // if we are using asynchronous send we should wait until the last send is complete
			MPI_Wait(&sendRequest, MPI_STATUS_IGNORE);

		gettimeofday(&zstpbegin, NULL);

		// Z-STEP BEGINS
		HforwardSGD(Xin, hX, N, L, Din);						// calc the output of updated hash functions
		ZsteprelaxedBUI(Xout, hX, Z, N, Dout, L, mu);			// update auxiliary variables using the relaxed solution
		ZstepalternateBUI(Xout, hX, Z, N, Dout, L, mu, 1, 0);	// update auxiliary variables by alternating over bits
		// Z-STEP ENDS

		gettimeofday(&zstpend, NULL);
		zstptime[iter] = HighResDiff(zstpend, zstpbegin);

		gettimeofday(&iterend, NULL);
		itertime[iter] = HighResDiff(iterend, iterbegin);

		int exportmodel = 1; // if set to 1 the model will be saved in a file at the end of each iteration
		if (mpirank == ROOT_PROCESS && exportmodel) // saving the model begins
		{
			int imodel;
			int imodelbuffindex = 0;
			for (imodel = 0; imodel < totaltaskcount; imodel++)
				if (imodel < L)
					getHmodel(imodel, modelbuffer, &imodelbuffindex);
				else
					getFmodel(imodel - L, modelbuffer, &imodelbuffindex);

			if (imodelbuffindex * sizeof(double) != modelsize)
				printf("inconsistent model %d %d\n", imodelbuffindex * (int)sizeof(double), modelsize);
			else
			{
				mkdir("model/", S_IRWXO | S_IRWXG | S_IRWXU);

				char *savepath = NULL;
				size_t size;
				FILE *stream;
				stream = open_memstream (&savepath, &size);
				fprintf (stream, "model/model%02d", iter + 1);
				fflush (stream);
				fclose (stream);

				fwritebuffer(savepath, (unsigned char*)modelbuffer, modelsize);
			}
		} // saving the model ends

		// calculate errors if requested by user at the end of each iteration
		if (report_inter_error || (report_final_error && iter == maxit - 1))
		{
			Herror(hX, Z, &(Herr[iter]), N, L);
			Ferror(Z, Xout, N, Dout, L, &(Ferr[iter]));
			Ferror(hX, Xout, N, Dout, L, &(Nestederr[iter]));

			int imodel;
			int imodelbuffindex = 0;
			for (imodel = 0; imodel < totaltaskcount; imodel++)
				if (imodel < L)
					getHmodel(imodel, modelbuffer, &imodelbuffindex);
				else
					getFmodel(imodel - L, modelbuffer, &imodelbuffindex);

			// DEBUG BEGIN
			if (imodelbuffindex * sizeof(double) != modelsize)
				printf("inconsistent model %d %d\n", imodelbuffindex * (int)sizeof(double), modelsize);
			else
			{
				//unsigned int checksum = fletcher32((unsigned short int*)modelbuffer, modelsize / sizeof(unsigned short int));
				//printf("P%02d model checksum: %010d\n", mpirank, checksum);
			}
			// DEBUG END

			// we sum up all the local errors using MPI_Allreduce to get the global errors and print them
			double totalHerr, totalFerr, totalNestederr;
			MPI_Allreduce(&(Herr[iter]), &totalHerr, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			MPI_Allreduce(&(Ferr[iter]), &totalFerr, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			MPI_Allreduce(&(Nestederr[iter]), &totalNestederr, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
			double totalobjective = totalFerr + mu * totalHerr;
			if (mpirank == ROOT_PROCESS)
			{
				printf("global error: obj-err = %12.8f h-err = %12.8f f-err = %12.8f\nglobal nested error: %12.8f\n",
						totalobjective, totalHerr, totalFerr, totalNestederr);
				TOTALobjecterr[iter] = totalobjective;
				TOTALnestederr[iter] = totalNestederr;
			}
		}

		if (report_inter_precision || (report_final_precision && iter == maxit - 1)) // calc the precision if requested
			TOTALprecision[iter + 1] = calcprecision(config, 0);
	} // end of main loop over iterations

	// DEBUG BEGIN
	int tindex, visits = ringepochs * maxit;
	for (tindex = 0; tindex < totaltaskcount; tindex++)
		if (taskvisits[tindex] != visits) // if any number of visits is different from expected print a warning
			printf("P%02d task%03d expected visits: %d visits: %d\n", mpirank, tindex, visits, taskvisits[tindex]);
	// DEBUG END

	double itersum, commsum, wstpsum, zstpsum;
	if (!synchronizeprocesses()) // calculate all the times and print
	{
		double totalitertime, totalcommtime, totalwstptime, totalzstptime;
		totalitertime = totalcommtime = totalwstptime = totalzstptime = 0;

		for (i = 0; i < maxit; i++)
		{
			totalitertime += itertime[i];
			totalcommtime += commtime[i];
			totalwstptime += wstptime[i];
			totalzstptime += zstptime[i];
		}

		//double unknowntime = totalitertime - totalcommtime - totalwstptime
		//		- totalzstptime;
		/*
		printf(
				"P%02d times. iter:%8.4f comm:%8.4f wstp:%8.4f zstp:%8.4f ?:%8.4f secs.\n",
				mpirank, totalitertime, totalcommtime, totalwstptime,
				totalzstptime, unknowntime);
		 */

		// get the total times by suming up using MPI_Allreduce
		MPI_Allreduce(&totalitertime, &itersum, 1, MPI_DOUBLE, MPI_SUM,	MPI_COMM_WORLD);
		MPI_Allreduce(&totalcommtime, &commsum, 1, MPI_DOUBLE, MPI_SUM,	MPI_COMM_WORLD);
		MPI_Allreduce(&totalwstptime, &wstpsum, 1, MPI_DOUBLE, MPI_SUM,	MPI_COMM_WORLD);
		MPI_Allreduce(&totalzstptime, &zstpsum, 1, MPI_DOUBLE, MPI_SUM,	MPI_COMM_WORLD);
	}

	synchronizeprocesses();

	if (mpirank == ROOT_PROCESS) // print all the errors and durations
	{
		printf("all processes ended successfully. duration: %12.8f comm: %12.8f wstp: %12.8f zstp: %12.8f\n",
				itersum / mpisize, commsum / mpisize, wstpsum / mpisize, zstpsum / mpisize);

		printf("\niteration time:\n");
		for (i = 0; i < maxit; i++)	printf("%12.8f ", itertime[i]);
		printf("\ncommunication time:\n");
		for (i = 0; i < maxit; i++)	printf("%12.8f ", commtime[i]);
		printf("\nw step time:\n");
		for (i = 0; i < maxit; i++)	printf("%12.8f (%12.8f,%12.8f) ", wstptime[i], wstptime_h[i], wstptime_f[i]);
		printf("\nz step time:\n");
		for (i = 0; i < maxit; i++)	printf("%12.8f ", zstptime[i]);

		printf("\nobjective errors:\n");
		for (i = 0; i < maxit; i++)
			if (TOTALobjecterr[i] >= 0) printf("%12.8f ", TOTALobjecterr[i]);
		printf("\nnested errors:\n");
		for (i = 0; i < maxit; i++)
			if (TOTALnestederr[i] >= 0) printf("%12.8f ", TOTALnestederr[i]);
		printf("\nprecisions:\n");
		for (i = 0; i <= maxit; i++)
			if (TOTALprecision[i] >= 0) printf("%f ", TOTALprecision[i]);
	}

	if (useBsend) // if we have used buffered MPI send
	{
		MPI_Buffer_detach(&mpi_attach_buff, &mpi_attach_buff_size); // detach the buffer from MPI environment
		free(mpi_attach_buff); // free the allocated memory
	}

	MPI_Finalize(); // finilize and close the MPI environment
	return EXIT_SUCCESS;
}
