// standard C libraries
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

// libraries used in GSL calculations
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_linalg.h>

// main header for any MPI program
#include "mpi.h"

#include "biguint.h"

// force GSL to access matrix elements as inline functions
#ifndef  HAVE_INLINE
#define  HAVE_INLINE
#endif

// different types of activation functions in the decoder layer
#define LAYER_TYPE_SIGMOID 1
#define LAYER_TYPE_LINEAR 2
#define LAYER_TYPE_SOFTMAX 3

// a unique number used as the "tag" for all the MPI messages carrying submodels
#define MODEL_MSG_TAG 19462874

// structure representing a single ANN layer
typedef struct
{
    int type;      // layer activation function type
    double l;      // (nonnegative scalar) regularisation parameter. default is 0.
    int units;     // number of units in layer; = Dout
    gsl_matrix* W; // (Din+1)*Dout weight matrix
}deepnet_layer;

// This struct is used to store the configuration settings for the program and
// the parameters are being read and set in the beginning of process by loading from a
// .conf text file. There is only one instance of it in the whole program.
typedef struct
{
// DATASET PARAMETERS ------------------------------------------------------------------------------------------------ 
	int L;				// size of binary codes: number of bits
	int Din;			// dimension of the input data in autoencoder, in the case of using
						// RBF input Din can be different from Dout. If this value is set to
						// zero the program will set it equal to Dout and use output datapoint
						// for input too.
	int Dout;			// dimension of the output data in autoencoder
	long int N;			// the actual size of input datapoints, used to calc presicion
	long int Ntrain;	// number of datapoints in the train dataset
	long int Ntest;		// number of datapoints in the test dataset

// PARMAC PARAMETERS -------------------------------------------------------------------------------------------------
	long int minibatchcnt;	// total number of minibatches, used to decide the minibatch size
	int epochs;			// number of epochs in W-step
	int maxit;			// number of iterations
	double *mu; 		// array containing the mu values. should be the size of maxit
	int K;				// neighbors in the original feature set
	int k;				// neighbors in the latent feature set
	double rbfsigma;	// the sigma used to calculate RBF inputs, is used only in case of
						// ONLINE RBF mode
// DATASET PATH ------------------------------------------------------------------------------------------------------ 
	char trainset[256];		// path of train dataset, used as output dataset
	char trainsetRBF[256];	// path RBF train dataset, used as input dataset, if this
							// value is "ONLINE" the program will use output dataset and
							// rbfsigma to generate the RBF input in place.
	char testset[256];		// path of the test dataset
	char neighbors[256];	// neighborhood matrix
	char initbintrain[256];	// initial train binary auxiliary values
	char initbintest[256];	// initial test binary auxiliary values
	char trainlabels[256];	// train labels, NOT USED
	char testlabels[256];	// test labels, NOT USED
	char rbfcenters[256];	// path to RBF centers, only used in case of ONLINE RBF mode

// DEBUG & FUNCTIONALITY ---------------------------------------------------------------------------------------------
	int initbinL;			// the original dimension of initial binary values, if the
							// binary values are coming from tPCA, we can simply use the
							// first L bits of this dataset for smaller L values.
	int reporterrs;			// a flag set by user to choose the items shown in output
							// values for this flag can be one of the followings:
									// 0: no reports
									// 1: just final results
									// 2: all
									// 3: all errors + final precision
									// 4: load models and calc ranks for recall@R
									// 5: load models and calc reconstruction error
									// 6: load models and calc both ranks and error

}configstruct;

// define the root process, it should be smaller than the total number of processes
#define ROOT_PROCESS 0
// inline definition to calc the time difference between two recorded times
#define HighResDiff(X,Y) ((X.tv_sec - Y.tv_sec) + (double)(X.tv_usec - Y.tv_usec) / 1000000)

// functions used in linear layer
void linearforward(gsl_matrix* X, deepnet_layer* f, gsl_matrix* fX, gsl_matrix* J);
void lineartrain(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, gsl_matrix* fX, double* fiterr, double* regerr);
void linearerror(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, double* err, gsl_matrix* grad);
void linearerrorWRAPPER(gsl_matrix* W, void* arglist[], double* err, gsl_matrix* grad);
void lineartrainCG(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, int maxit, gsl_matrix* fX, double* fiterr, double* regerr);
void lineartrainCG_byunit(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, int maxit, gsl_matrix* fX, double* fiterr, double* regerr);
void lineartrainSD(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, double step, gsl_matrix* fX, double* fiterr, double* regerr);

// minimize optimization function, translated from MATLAB source code
void minimize (gsl_matrix* X, void (*f)(gsl_matrix*, void* [], double*, gsl_matrix*), int length, void* arglist[], double* fX, int* i);

// general ANN functions
void deepnetforward(gsl_matrix* X, deepnet_layer* F[], int Nlayers, gsl_matrix* fX, gsl_matrix* Y, double* fiterr);
void layerforward(gsl_matrix* X, deepnet_layer* f, gsl_matrix* fX, gsl_matrix* J);
void layererror(gsl_matrix* Y, deepnet_layer* f, gsl_matrix* fX, double* fiterr, double* regerr);

// Z-step functions
void Zinit(long int N, int Dx, int L);
void Zstepexact(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu);
void Zstepfastexact(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu);
void Zstepalternate(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int blocksize, int maxit);
void Zstepalternate_1(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int maxit);
void Zsteprelaxed(gsl_matrix *X, gsl_matrix *hX, gsl_matrix *Z, double mu, double r);
void ZsteprelaxedBUI(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu);
void ZstepalternateBUI(outype *X, biguint *hX, biguint *Z, long int N, int Dx, int L, double mu, int blocksize, int maxit);

// H-step functions
//void Hinit(gsl_matrix *X, int L);
//void Hstep(gsl_matrix *Z, gsl_matrix *hX, double *Herr);
void Herror(biguint *hX, biguint *Z, double *Herr, long int N, int L);
void HinitSGD(int Dx, int L);
void HtrainSGD(intype *X, biguint *hX, long int N, int Dx, int chunks, int epochs, int isvm, int stepcounter, int *minibatchperm);
void HforwardSGD(intype *X, biguint *hX, long int N, int L, int Dx);
void getHmodel(int isvm, double *buffer, int *index);
void setHmodel(int isvm, double *buffer, int *index);

// F-step functions
void Finit(int L, int Dx, int breakf);
void Ftrain(gsl_matrix *Z, gsl_matrix *X);
void Fforward(gsl_matrix *X, gsl_matrix *fX);
void Ferror(biguint *Z, outype *X, long int N, int Dx, int L, double *Ferr);
void FtrainCG(biguint *Z, outype *X, long int N, int Dx, int L, int chunks, int epochs);
void FtrainSGD(biguint *Z, outype *X, long int N, int Dx, int L, int chunks, int epochs, int iann, int stepcounter, int *minibatchperm);
void FtrainexactSGD(gsl_matrix *Z, gsl_matrix *X, int chunks, int epochs, double alpha, int iann, int reset);
void getFmodel(int iann, double *buffer, int *index);
void setFmodel(int iann, double *buffer, int *index);

// debug and test functions
void Ftest();
void Ztest();
void fastKNN(int datapoints, int bits);

// utility functions
long int loaddatasets(configstruct *config, intype **Xin, outype **Xout, biguint **Z,int mpirank, int mpisize, int loadbinaries);
void print_matrix(char* name, gsl_matrix* m);
int synchronizeprocesses();
int** dividetasks(int tasks, int processors, int *taskgroups);
int* calcbuffersizes(int Din, int Dout, int L, int taskgroups, int **tasks, int breakf);
unsigned int fletcher32(unsigned short int *data, int words);
void freadmatrix(char *filename, gsl_matrix *m);
void freadbuffer(char *filename, unsigned char *buffer, long int len, long int offset);
void fwritebuffer(char *filename, unsigned char *buffer, int len);
double calcprecision(configstruct *config, int loadbinaries);
configstruct* loadconfig(char *path, int report);
int* loadperm(char *path, int i, int e, int p, long int m, int report);
void calcranks(biguint *test, long int testN, biguint *query, long int queryN, biguint *nn, int *ranks, int worstcase);
void calcgndbins(configstruct *config, unsigned int *gnd, biguint *gndbins, int loadbinaries);
void calcRBF(outype *Xout, int N, int Dout, outype *R, int rbcount, double sigma, intype *Xin);
int process_trained_models(int calculate_ranks, int calculate_AEerr, configstruct *config,
		int mpirank, double *modelbuffer, int modelsize, int totaltaskcount,
		intype *Xin, outype *Xout, biguint *Z, long int N, double *Nestederr, double *TOTALnestederr);


