#include <stdint.h>
#include "BA.h"

// in order to be able to apply efficient bit operations on binary numbers of any size
// we have defined this new struct which represents a big unsigned int, also the required
// functions are implemented for this new datatype

//types and constants used in the functions below
const uint64_t m1  = 0x5555555555555555; //binary: 0101...
const uint64_t m2  = 0x3333333333333333; //binary: 00110011..
const uint64_t m4  = 0x0f0f0f0f0f0f0f0f; //binary:  4 zeros,  4 ones ...
const uint64_t m8  = 0x00ff00ff00ff00ff; //binary:  8 zeros,  8 ones ...
const uint64_t m16 = 0x0000ffff0000ffff; //binary: 16 zeros, 16 ones ...
const uint64_t m32 = 0x00000000ffffffff; //binary: 32 zeros, 32 ones
const uint64_t hff = 0xffffffffffffffff; //binary: all ones
const uint64_t h01 = 0x0101010101010101; //the sum of 256 to the power of 0,1,2,3...

//This is a naive implementation, shown for comparison,
//and to help in understanding the better functions.
//It uses 24 arithmetic operations (shift, add, and).
inline int popcount_1(uint64_t x) {
    x = (x & m1 ) + ((x >>  1) & m1 ); //put count of each  2 bits into those  2 bits
    x = (x & m2 ) + ((x >>  2) & m2 ); //put count of each  4 bits into those  4 bits
    x = (x & m4 ) + ((x >>  4) & m4 ); //put count of each  8 bits into those  8 bits
    x = (x & m8 ) + ((x >>  8) & m8 ); //put count of each 16 bits into those 16 bits
    x = (x & m16) + ((x >> 16) & m16); //put count of each 32 bits into those 32 bits
    x = (x & m32) + ((x >> 32) & m32); //put count of each 64 bits into those 64 bits
    return x;
}

//This uses fewer arithmetic operations than any other known
//implementation on machines with slow multiplication.
//It uses 17 arithmetic operations.
inline int popcount_2(uint64_t x) {
    x -= (x >> 1) & m1;             //put count of each 2 bits into those 2 bits
    x = (x & m2) + ((x >> 2) & m2); //put count of each 4 bits into those 4 bits
    x = (x + (x >> 4)) & m4;        //put count of each 8 bits into those 8 bits
    x += x >>  8;  //put count of each 16 bits into their lowest 8 bits
    x += x >> 16;  //put count of each 32 bits into their lowest 8 bits
    x += x >> 32;  //put count of each 64 bits into their lowest 8 bits
    return x & 0x7f;
}

//This uses fewer arithmetic operations than any other known
//implementation on machines with fast multiplication.
//It uses 12 arithmetic operations, one of which is a multiply.
inline int popcount_3(uint64_t x) {
    x -= (x >> 1) & m1;             //put count of each 2 bits into those 2 bits
    x = (x & m2) + ((x >> 2) & m2); //put count of each 4 bits into those 4 bits
    x = (x + (x >> 4)) & m4;        //put count of each 8 bits into those 8 bits
    return (x * h01)>>56;  //returns left 8 bits of x + (x<<8) + (x<<16) + (x<<24) + ...
}

//This is better when most bits in x are 0
//It uses 3 arithmetic operations and one comparison/branch per "1" bit in x.
inline int popcount_4(uint64_t x) {
    int count;
    for (count=0; x; count++)
        x &= x-1;
    return count;
}

// calculates the hamming distance between to biguints using the builtin popcount function of the compiler
// if the hardware instruction for popcount is avaialbe it will be used, otherwise it will be replaces with
// some other software implementation
inline int bui_fasthamming_builtin(biguint *bui1, biguint *bui2)
{
	int hamming = 0;
	int i;
	for (i = 0; i < bui_blen; i++)
		hamming += __builtin_popcount(bui1->b[i] ^ bui2->b[i]);
	return hamming;
}

// a very efficient hamming distance calculator for 64 bit values
inline int bui_fasthammingx64(biguint *bui1, biguint *bui2)
{
	// we know biguint is 64 bits
	uint64_t x = bui1->b[0] ^ bui2->b[0];
	x = (x << 32) + (bui1->b[1] ^ bui2->b[1]);
	return popcount_3(x);
}

// set to zero
inline void bui_zero(biguint *bui)
{
	int i;
	for (i = 0; i < bui_blen; i++)
		bui->b[i] = 0;
}

// set the value of a single bit by it's index
inline void bui_setbit(biguint *bui, int bit, int val)
{
	/*
	int block = bit / (sizeof(unsigned int) * 8);
	bit %= sizeof(unsigned int) * 8;
	unsigned int one = 1;

	if (val)
		bui->b[block] |= one << bit;
	else
		bui->b[block] &= ~(one << bit);
	*/

	if (val)
		bui->b[bit / 32] |= 1 << (bit % 32);
	else
		bui->b[bit / 32] &= ~(1 << ((bit % 32)));
}

// check if the two biguints have the same value
inline int bui_equal(biguint *bui1, biguint *bui2) // 0:bui1!=bui2 1:bui1==bui2
{
	int i;
	for (i = 0; i < bui_blen; i++)
		if (bui1->b[i] != bui2->b[i])
			return 0;
	return 1;
}

// increase the value of a given biguint by 1
inline void bui_inc(biguint *bui)
{
	int block = 0;
	bui->b[block] += 1;

	while (bui->b[block] == 0)
	{
		block++;
		if (block == bui_blen)
			break;
		bui->b[block] += 1;
	}
}

// get the value of a single bit by it's index
inline int bui_getbit(biguint *bui, int bit)
{
	/*
	int block = bit / (sizeof(unsigned int) * 8);
	bit %= sizeof(unsigned int) * 8;
	unsigned int one = 1;

	return (bui->b[block]) & (one << bit) ? 1 : 0;
	*/

	return (bui->b[bit / 32]) & (1 << (bit % 32)) ? 1 : 0;
}

// set the value of one biguint equal to another biguint variable
inline void bui_set(biguint *bui1, biguint *bui2)
{
	int i;
	for (i = 0; i < bui_blen; i++)
		bui1->b[i] = bui2->b[i];
}
