#include "BA.h"

/*
  linearforward: computes outputs of a linear layer f(x) = x * W + b and its Jacobian.

  In:
  X: N*Din matrix, N Din-dim data points rowwise.
  f: The linear layer struct.
  Out:
  fX: N*Dout matrix, N Dout-dim outputs f(X).
  J: Dout*Din Jacobian matrix.

  Copyright (c) 2014 by Miguel A. Carreira-Perpinan, Weiran Wang and Mehdi Alizadeh.
*/

void linearforward(gsl_matrix* X, deepnet_layer* f, gsl_matrix* fX, gsl_matrix* J)
{
    int N = X->size1;
    int Din = f->W->size1 - 1; // The last row is bias
    int Dout = f->W->size2;
  
    gsl_matrix_view weights = gsl_matrix_submatrix(f->W, 0, 0, Din, Dout);
    gsl_matrix_view bias = gsl_matrix_submatrix(f->W, Din, 0, 1, Dout);

    // Compute Y = X * weights
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, X, &(weights.matrix), 0.0, fX);

    // Add bias to Y
    int i;
    for(i = 0; i < N; i++)
    {
		gsl_matrix_view fXi = gsl_matrix_submatrix(fX, i , 0 , 1, Dout);
		gsl_matrix_add(&(fXi.matrix), &(bias.matrix));
    }

    // Jacobian matrix is just the transpose of weights in a linear layer
    if (J)
    {
    	gsl_matrix_transpose_memcpy(J, &(weights.matrix));
    }
}

/*
  lineartrain: trains weights of a linear layer.

  In:
  X: N*Din matrix, N Din-dim data points rowwise.
  Y: N*Dout matrix, N Dout-dim data points rowwise.
  l: (nonnegative scalar) regularisation parameter. // default is 0
  Out:
  f: (struct) the linear function, with fields:
  units = Dout, type = LAYER_TYPE_LINEAR, (Din+1)*Dout weight matrix, regularisation parameter l.
  fX: N*Dout matrix, f(X).
  fiterr: fit error.
  regerr: regularisation error.

  Copyright (c) 2014 by Miguel A. Carreira-Perpinan, Weiran Wang and Mehdi Alizadeh.
*/

void lineartrain(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, gsl_matrix* fX, double* fiterr, double* regerr)
{
    int N = X->size1;
    int Din = X->size2;
    int Dout = Y->size2;

    // To reduce memory cost, create views of weights and bias from W matrix.
    gsl_matrix_view v1 = gsl_matrix_submatrix(f->W, 0, 0, Din, Dout);
    gsl_matrix* weights = &(v1.matrix);
    gsl_matrix_view v2 = gsl_matrix_submatrix(f->W,Din,0,1,Dout);
    gsl_matrix* bias = &(v2.matrix);  

    // Compute X1 = X' * ones.
    gsl_matrix* ones = gsl_matrix_calloc(N, 1);
    gsl_matrix_set_all(ones, 1);
    gsl_matrix* X1 = gsl_matrix_calloc(Din, 1);
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, X, ones, 0.0, X1);

    // Compute XX = X' * X - X1 * (X1' / N).
    gsl_matrix* XX = gsl_matrix_calloc(Din, Din);
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, X, X, 0.0, XX);
    gsl_blas_dgemm(CblasNoTrans, CblasTrans, -1.0 / N, X1, X1, 1.0, XX);
    
    // Add regularization.
    if (f->l > 0)
    {
    	gsl_vector_view v = gsl_matrix_diagonal(XX);
    	gsl_vector_add_constant(&(v.vector), f->l);
    }

    // Compute (X' * Y - X1 * mean(Y, 1)).
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, X, Y, 0.0, weights);
    gsl_matrix * Y1 = gsl_matrix_calloc(1,Dout);
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, ones, Y, 0.0, Y1);
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, -1.0/N, X1, Y1, 1.0, weights);

    // XX is now only used to solve linear systems, so Cholesky decomposition can be in-place.
    gsl_linalg_cholesky_decomp(XX);

    // Compute weights: weights = XX \ (X' * Y - X1 * mean(Y, 1)).
    gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1.0, XX, weights);
    gsl_blas_dtrsm(CblasLeft, CblasUpper, CblasNoTrans, CblasNonUnit, 1.0, XX, weights);
    gsl_matrix_free(XX);

    // Compute bias = mean(Y - X * weights, 1).
    gsl_matrix* temp = gsl_matrix_calloc(N, Dout);
    gsl_matrix_memcpy(temp, Y);
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, -1.0, X, weights, 1.0, temp); 
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0 / N, ones, temp, 0.0, bias);

    gsl_matrix_free(ones);
    gsl_matrix_free(X1);
    gsl_matrix_free(Y1);
    gsl_matrix_free(temp);

    if (fX)
    {
    	linearforward(X, f, fX, NULL);
    	layererror(Y, f, fX, fiterr, regerr);
    }
}

void linearerror(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, double* err, gsl_matrix* grad)
{
    int N = X->size1;
    int Din = f->W->size1 - 1; // Last row is bias
    int Dout = f->W->size2;
  
    gsl_matrix_view weights = gsl_matrix_submatrix(f->W, 0, 0, Din, Dout);
    gsl_matrix_view bias = gsl_matrix_submatrix(f->W, Din, 0, 1, Dout);

    gsl_matrix* fX = gsl_matrix_calloc(N, Dout);
  
    // Compute fX = X * weights
    gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, X, &(weights.matrix), 0.0, fX);

    // Add the bias to fX
    int i, j;
    for(i = 0; i < N; i++)
    {
    	gsl_matrix_view fXi = gsl_matrix_submatrix(fX, i, 0, 1, Dout);
    	gsl_matrix_add(&(fXi.matrix), &(bias.matrix));
    }

    gsl_matrix* diff = gsl_matrix_calloc(N, Dout);
    gsl_matrix_memcpy(diff, fX);
    gsl_matrix_sub(diff, Y);
 
    double t; 
    // Compute error between target Y and output fX.
    *err = 0;
    for(i = 0; i < N; i++)
    	for(j = 0; j < Dout; j++)
    	{
    		//t = gsl_matrix_get(diff, i, j);
    		t = diff->data[i * Dout + j];
    		(*err) += t * t;
    	}

    // Add regularization.
    for(i = 0; i < Din; i++)
		for(j = 0; j < Dout; j++)
		{
			//t = gsl_matrix_get(f->W, i, j);
			t = f->W->data[i * Dout + j];
			(*err) += f->l * t * t;
		}
    
	//printf("err: %f\n", *err);

    if (grad)
    {
		// Overwrite diff.
		gsl_matrix_scale(diff, 2.0);

		gsl_matrix* ones = gsl_matrix_calloc(N, 1);

		// Compute gradient.
		gsl_matrix_set_zero(grad);

		gsl_matrix_view gv1 = gsl_matrix_submatrix(grad, 0, 0, Din, Dout);
		gsl_matrix_memcpy(&(gv1.matrix), &(weights.matrix));
		gsl_matrix_scale(&(gv1.matrix), 2 * f->l);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, X, diff, 1.0, &(gv1.matrix));

		gsl_matrix_view gv2 = gsl_matrix_submatrix(grad, Din, 0, 1, Dout);
		gsl_matrix_set_all(ones, 1);
		gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, ones, diff, 0.0, &(gv2.matrix));

		gsl_matrix_free(ones);
    }

    gsl_matrix_free(fX);
    gsl_matrix_free(diff);
}

void linearerrorWRAPPER(gsl_matrix* W, void* arglist[], double* err, gsl_matrix* grad)
{
    gsl_matrix* X = (gsl_matrix*)arglist[0];
    gsl_matrix* Y = (gsl_matrix*)arglist[1];
    deepnet_layer* f = (deepnet_layer*)arglist[2];

    gsl_matrix *Wtemp = f->W;
    f->W = W;
    linearerror(X, Y, f, err, grad);
    f->W = Wtemp;
}

void lineartrainCG(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, int maxit, gsl_matrix* fX, double* fiterr, double* regerr)
{
    void* arglist[3] = {X, Y, f};
    double E[maxit + 1];
    int itCount;

    minimize(f->W, &linearerrorWRAPPER, maxit, arglist, E, &itCount);

    if (fX)
    {
    	//struct timeval comm_begin;
		//struct timeval comm_end;

		linearforward(X, f, fX, NULL);
		layererror(Y, f, fX, fiterr, regerr);
    }
}

void lineartrainCG_byunit(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, int maxit, gsl_matrix* fX, double* fiterr, double* regerr)
{
    void* arglist[3] = {X, Y, f};
    double E[maxit + 1];
    int itCount;

    int N = Y->size1;
    int Din = X->size2;
    int Dout = Y->size2;
    int j;

    for(j = 0; j < Dout; j++)
    {
		gsl_matrix_view yv = gsl_matrix_submatrix(Y, 0, j, N, 1);
		gsl_matrix_view wv = gsl_matrix_submatrix(f->W, 0, j, Din + 1, 1);

		arglist[1] = &(yv.matrix);

		minimize(&(wv.matrix), &linearerrorWRAPPER, maxit, arglist, E, &itCount);
    }

    if (fX)
    {
		//struct timeval comm_begin;
		//struct timeval comm_end;

		linearforward(X, f, fX, NULL);
		layererror(Y, f, fX, fiterr, regerr);
    }
}

void lineartrainSD(gsl_matrix* X, gsl_matrix* Y, deepnet_layer* f, double step, gsl_matrix* fX, double* fiterr, double* regerr)
{
    int Din = X->size2;
    int Dout = Y->size2;
	gsl_matrix* grad = gsl_matrix_calloc(Din + 1, Dout);

	double err;
	linearerror(X, Y, f, &err, grad);

	int N = X->size1;
	if (err > N * 1000)
		printf("big error: %f\n", err);

	gsl_matrix_scale(grad, -step);
	gsl_matrix_add(f->W, grad);

    if (fX)
    {
    	//struct timeval comm_begin;
		//struct timeval comm_end;

		linearforward(X, f, fX, NULL);
		layererror(Y, f, fX, fiterr, regerr);
    }

    gsl_matrix_free(grad);
}
