#include "BA.h"

/*
  % Minimize a differentiable multivariate function.
  %
  % Usage: [X, fX, i] = minimize(X, f, length, P1, P2, P3, ... )
  %
  % where the starting point is given by "X" (D by 1), and the function named in
  % the string "f", must return a function value and a vector of partial
  % derivatives of f wrt X, the "length" gives the length of the run: if it is
  % positive, it gives the maximum number of line searches, if negative its
  % absolute gives the maximum allowed number of function evaluations. You can
  % (optionally) give "length" a second component, which will indicate the
  % reduction in function value to be expected in the first line-search (defaults
  % to 1.0). The parameters P1, P2, P3, ... are passed on to the function f.
  %
  % The function returns when either its length is up, or if no further progress
  % can be made (ie, we are at a (local) minimum, or so close that due to
  % numerical problems, we cannot get any closer). NOTE: If the function
  % terminates within a few iterations, it could be an indication that the
  % function values and derivatives are not consistent (ie, there may be a bug in
  % the implementation of your "f" function). The function returns the found
  % solution "X", a vector of function values "fX" indicating the progress made
  % and "i" the number of iterations (line searches or function evaluations,
  % depending on the sign of "length") used.
  %
  % The Polack-Ribiere flavour of conjugate gradients is used to compute search
  % directions, and a line search using quadratic and cubic polynomial
  % approximations and the Wolfe-Powell stopping criteria is used together with
  % the slope ratio method for guessing initial step sizes. Additionally a bunch
  % of checks are made to make sure that exploration is taking place and that
  % extrapolation will not be unboundedly large.
  %
  % See also: checkgrad
  %
  % Copyright (C) 2001 - 2006 by Carl Edward Rasmussen (2006-09-08).
*/

void minimize (gsl_matrix* X, void (*f)(gsl_matrix*, void* [], double*, gsl_matrix*), int length, void* arglist[], double* fX, int* i)
{
    //double realmin = 2.225073858507201e-308; // To be consistent with MATLAB.
    double realmin = 1e-14;

    double INT = 0.1; // don't reevaluate within 0.1 of the limit of the current bracket
    double EXT = 3.0; // extrapolate maximum 3 times the current step-size
    double MAX = 20.0; // max 20 function evaluations per line search 
    double RATIO = 10.0; // maximum allowed slope ratio
    double SIG = 0.1; double RHO = SIG / 2; /* SIG and RHO are the constants controlling the Wolfe-
					       % Powell conditions. SIG is the maximum allowed absolute ratio between
					       % previous and new slopes (derivatives in the search direction), thus setting
					       % SIG to low (positive) values forces higher precision in the line-searches.
					       % RHO is the minimum allowed fraction of the expected (from the slope at the
					       % initial point in the linesearch). Constants must satisfy 0 < RHO < SIG < 1.
					       % Tuning of SIG (depending on the nature of the function to be optimized) may
					       % speed up the minimization; it is probably not worth playing much with RHO.
					    */
/*
  % The code falls naturally into 3 parts, after the initial line search is
  % started in the direction of steepest descent. 1) we first enter a while loop
  % which uses point 1 (p1) and (p2) to compute an extrapolation (p3), until we
  % have extrapolated far enough (Wolfe-Powell conditions). 2) if necessary, we
  % enter the second loop which takes p2, p3 and p4 chooses the subinterval
  % containing a (local) minimum, and interpolates it, unil an acceptable point
  % is found (Wolfe-Powell conditions). Note, that points are always maintained
  % in order p0 <= p1 <= p2 < p3 < p4. 3) compute a new search direction using
  % conjugate gradients (Polack-Ribiere flavour), or revert to steepest if there
  % was a problem in the previous line-search. Return the best value so far, if
  % two consecutive line-searches fail, or whenever we run out of function
  % evaluations or line-searches. During extrapolation, the "f" function may fail
  % either with an error or returning Nan or Inf, and minimize should handle this
  % gracefully.
*/

    double red = 1.0;

/*
  char* S;
  if (length > 0)
  S = "Linesearch";
  else
  S = "Function evaluation";
*/

    double f0, f1, f2, f3, f4; f0 = f1 = f2 = f3 = f4 = 0;
    double d0, d1, d2, d3, d4; d0 = d1 = d2 = d3 = d4 = 0;
    double /*x0,*/ x1, x2, x3, x4; /*x0 =*/ x1 = x2 = x3 = x4 = 0;

    int p, q; // loop counters

    int Xdim1 = X->size1;
    int Xdim2 = X->size2;

    gsl_matrix*  df0 = gsl_matrix_calloc(Xdim1, Xdim2);
    gsl_matrix* tmpX = gsl_matrix_calloc(Xdim1, Xdim2);
    gsl_matrix*  df3 = gsl_matrix_calloc(Xdim1, Xdim2);
    gsl_matrix*   X0 = gsl_matrix_calloc(Xdim1, Xdim2);
    gsl_matrix*  dF0 = gsl_matrix_calloc(Xdim1, Xdim2);
    gsl_matrix*    s = gsl_matrix_calloc(Xdim1, Xdim2);

    (*i) = 0;
    int ls_failed = 0;

    (*f)(X, arglist, &f0, df0);

    int fXcounter = 0;
    fX[fXcounter++] = f0;
    (*i) += (length < 0 ? 1 : 0);

// s = -df0
    gsl_matrix_memcpy(s, df0);
    gsl_matrix_scale(s, -1.0);

// d0 = -s'*s
    d0 = 0;
    for (p = 0; p < Xdim1; p++)
	for (q = 0; q < Xdim2; q++) 
	{
	    double t = gsl_matrix_get(s, p, q);
	    d0 += -t * t;
	}
    x3 = red / (1.0 - d0);

    while ((*i) < abs(length))
    {
	(*i) += length > 0 ? 1 : 0;

	gsl_matrix_memcpy(X0, X);
	double F0 = f0;
	gsl_matrix_memcpy(dF0, df0);
	int M = MAX;
	if (length <= 0)
	    M = MAX < -length - (*i) ? MAX : -length - (*i);
    
	while(1) 
	{
	    x2 = 0; f2 = f0; d2 = d0; f3 = f0;
	    gsl_matrix_memcpy(df3, df0);

	    int success = 0;

	    gsl_matrix_memcpy(tmpX, s);
	    gsl_matrix_scale(tmpX, x3);
	    gsl_matrix_add(tmpX, X);

	    while (!success && M > 0) 
	    {
		M--;
		(*i) += length < 0 ? 1 : 0;
		(*f)(tmpX, arglist, &f3, df3);
		// Numerical issues arise if any element of f3 and df3 is NaN or Inf.
		if (0 /* if any error occures inside f or outputs are not valid */) // needs further consideration
		    x3 = (x2 + x3) / 2.0;
		else
		    success = 1;
	    }

	    if (f3 < F0) 
	    {
		gsl_matrix_memcpy(X0, s);
		gsl_matrix_scale(X0, x3);
		gsl_matrix_add(X0, X);

		F0 = f3;
		gsl_matrix_memcpy(dF0, df3);
	    }

	    d3 = 0.0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		    d3 += gsl_matrix_get(s, p, q) * gsl_matrix_get(df3, p, q);

	    if (d3 > SIG * d0 || f3 > f0 + x3 * RHO * d0 || M == 0)
		break;

	    x1 = x2; f1 = f2; d1 = d2;
	    x2 = x3; f2 = f3; d2 = d3;

	    double A = 6.0 * (f1 - f2) + 3.0 * (d2 + d1) * (x2 - x1);
	    double B = 3.0 * (f2 - f1) - (2.0 * d1 + d2) * (x2 - x1);
      
	    double p1 = B * B - A * d1 * (x2 - x1);

	    if (p1 < 0)
		x3 = x2 * EXT;
	    else
	    {
		double p2 = B + sqrt(p1);
		if (p2 == 0)
		    x3 = x2 * EXT;
		else
		{
		    x3 = x1 - d1 * (x2 - x1) * (x2 - x1) / p2;
		    if (x3 < 0)
			x3 = x2 * EXT;  
		    else if (x3 > x2 * EXT)
			x3 = x2 * EXT;
		    else if (x3 < x2 + INT * (x2 - x1))
			x3 = x2 + INT * (x2 - x1);
		}
	    }
	}

	while ((fabs(d3) > -SIG * d0 || f3 > f0 + x3 * RHO * d0) && M > 0)
	{
	    if (d3 > 0 || f3 > f0 + x3 * RHO * d0)
	    {
		x4 = x3; f4 = f3; d4 = d3;
	    }
	    else 
	    {
		x2 = x3; f2 = f3; d2 = d3;
	    }

	    if (f4 > f0)
	    {
		double p1 = f4 - f2 - d2 * (x4 - x2);

		if (p1 == 0)
		    x3 = (x2 + x4) / 2.0;
		else
		    x3 = x2 - (0.5 * d2 * (x4 - x2) * (x4 - x2)) / p1;
	    }
	    else 
	    {
		double A = 6.0 * (f2 - f4) / (x4 - x2) + 3.0 * (d4 + d2);
		double B = 3.0 * (f4 - f2) - (2.0 * d2 + d4) * (x4 - x2);

		double p1 = B * B - A * d2 * (x4 - x2) * (x4 - x2);

		if (p1 < 0 || A == 0)
		    x3 = (x2 + x4) / 2.0;
		else
		    x3 = x2 + (sqrt(p1) - B) / A;
	    }

	    double p1 = x4 - INT * (x4 - x2);
	    p1 = x3 < p1 ? x3 : p1;
	    double p2 = x2 + INT * (x4 - x2);
	    x3 = p1 > p2 ? p1 : p2;
 
	    gsl_matrix_memcpy(tmpX, s);
	    gsl_matrix_scale(tmpX, x3);
	    gsl_matrix_add(tmpX, X);

	    (*f)(tmpX, arglist, &f3, df3);

	    if (f3 < F0)
	    {
		gsl_matrix_memcpy(X0,tmpX);
		F0 = f3;
		gsl_matrix_memcpy(dF0, df3);
	    }

	    M--;
	    (*i) += length < 0 ? 1 : 0;

	    d3 = 0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		    d3 += gsl_matrix_get(s, p, q) * gsl_matrix_get(df3, p, q);
	}

	if (fabs(d3) < -SIG * d0 && f3 < f0 + x3 * RHO * d0)
	{
	    gsl_matrix_memcpy(tmpX, s);
	    gsl_matrix_scale(tmpX, x3);
	    gsl_matrix_add(tmpX, X);
	    gsl_matrix_memcpy(X, tmpX);

	    f0 = f3;

	    fX[fXcounter++] = f0;

	    // s = (df3' * df3 - df0' * df3) / (df0' * df0) * s - df3;   % Polack-Ribiere CG direction
	    double p1 = 0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		{
		    double t = gsl_matrix_get(df3, p, q);
		    p1 += t * t;
		}

	    double p2 = 0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		    p2 += gsl_matrix_get(df0, p, q) * gsl_matrix_get(df3, p, q);

	    double p3 = 0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		{
		    double t = gsl_matrix_get(df0, p, q);
		    p3 += t * t;
		}

	    gsl_matrix_scale(s, (p1 - p2) / p3);
	    gsl_matrix_sub(s, df3);

	    gsl_matrix_memcpy(df0, df3);
	    d3 = d0;

	    d0 = 0;
	    for (p = 0; p < Xdim1; p++) 
		for (q = 0; q < Xdim2; q++) 
		    d0 += gsl_matrix_get(s, p, q) * gsl_matrix_get(df0, p, q);

	    if (d0 > 0) 
	    {
		gsl_matrix_memcpy(s, df0);
		gsl_matrix_scale(s, -1.0);

		d0 = 0;
		for (p = 0; p < Xdim1; p++) 
		    for (q = 0; q < Xdim2; q++) 
		    {
			double t = gsl_matrix_get(s, p, q);
			d0 += -t * t;
		    }
	    }

	    x3 = x3 * (RATIO < d3 / (d0 - realmin) ? RATIO : d3 / (d0 - realmin));
	    ls_failed = 0;
	}
	else 
	{
	    gsl_matrix_memcpy(X, X0);
	    f0 = F0;
	    gsl_matrix_memcpy(df0, dF0);

	    if (ls_failed || (*i) > abs(length))
		break;

	    gsl_matrix_memcpy(s, df0);
	    gsl_matrix_scale(s, -1.0);

	    d0 = 0;
	    for (p = 0; p < Xdim1; p++)
		for (q = 0; q < Xdim2; q++)
		{
		    double t = gsl_matrix_get(s, p, q);
		    d0 += -t * t;
		}

	    x3 = 1.0 / (1.0 - d0);
	    ls_failed = 1;
	}
    }
  
    gsl_matrix_free( df0);
    gsl_matrix_free( dF0);
    gsl_matrix_free( df3);
    gsl_matrix_free(   s);
    gsl_matrix_free(tmpX);
    gsl_matrix_free(  X0);
}
