// -*- C++ -*-
// SVM with stochastic gradient
// Copyright (C) 2007- Leon Bottou

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA

#include <iostream>
#include <iomanip>
#include <string>
#include <map>
#include <vector>
#include <cstdlib>
#include <cmath>

#include "../../biguint.h"

#include "lib/assert.h"
#include "lib/vectors.h"
#include "loss.h"
#include "svmsgd.h"

using namespace std;

// ---- Loss function

// Compile with -DLOSS=xxxx to define the loss function.
// Loss functions are defined in file loss.h)
#ifndef LOSS
# define LOSS HingeLoss
#endif

// ---- Bias term

// Compile with -DBIAS=[1/0] to enable/disable the bias term.
// Compile with -DREGULARIZED_BIAS=1 to enable regularization on the bias term

#ifndef BIAS
# define BIAS 1
#endif
#ifndef REGULARIZED_BIAS
# define REGULARIZED_BIAS 0
#endif

// ---- Plain stochastic gradient descent

class SvmSgd
{
public:
  SvmSgd(int dim, double lambda, double eta0  = 0);
  void renorm();
  double wnorm();
  double testOne(const XVector &x, double y, double *ploss, double *pnerr, double *pout = NULL);
  void trainOne(const XVector &x, double y, double eta);
public:
  void train(int imin, int imax, const xvec_t &x, const yvec_t &y, const char *prefix = "");
  void test(int imin, int imax, const xvec_t &x, const yvec_t &y, const char *prefix = "");
public:
  double evaluateEta(int imin, int imax, const xvec_t &x, const yvec_t &y, double eta);
  void determineEta0(int imin, int imax, const xvec_t &x, const yvec_t &y);
//private: // removed private to be able to set easily from outside! :D
  double  lambda;
  double  eta0;
  FVector w;
  double  wDivisor;
  double  wBias;
  double  t;
};

/// Constructor
SvmSgd::SvmSgd(int dim, double lambda, double eta0)
  : lambda(lambda), eta0(eta0), 
    w(dim), wDivisor(1), wBias(0),
    t(0)
{
	//cout << "SvmSgd Constructor reached." << endl;
}

/// Renormalize the weights
void
SvmSgd::renorm()
{
  if (wDivisor != 1.0)
    {
      w.scale(1.0 / wDivisor);
      wDivisor = 1.0;
    }
}

/// Compute the norm of the weights
double
SvmSgd::wnorm()
{
  double norm = dot(w,w) / wDivisor / wDivisor;
#if REGULARIZED_BIAS
  norm += wBias * wBias
#endif
  return norm;
}

/// Compute the output for one example.
double
SvmSgd::testOne(const XVector &x, double y, double *ploss, double *pnerr, double *pout)
{
  double s = dot(w,x) / wDivisor + wBias;
  if (ploss)
    *ploss += LOSS::loss(s, y);
  if (pnerr)
    *pnerr += (s * y <= 0) ? 1 : 0;
  if (pout)
	*pout = s;
  return s;
}

/// Perform one iteration of the SGD algorithm with specified gains
void
SvmSgd::trainOne(const XVector &x, double y, double eta)
{
  double s = dot(w,x) / wDivisor + wBias;
  // update for regularization term
  wDivisor = wDivisor / (1 - eta * lambda);
  if (wDivisor > 1e5) renorm();
  // update for loss term
  double d = LOSS::dloss(s, y);
  if (d != 0)
    w.add(x, eta * d * wDivisor);
  // same for the bias
#if BIAS
  double etab = eta * 0.01;
#if REGULARIZED_BIAS
  wBias *= (1 - etab * lambda);
#endif
  wBias += etab * d;
#endif
}


/// Perform a training epoch
void 
SvmSgd::train(int imin, int imax, const xvec_t &xp, const yvec_t &yp, const char *prefix)
{
  //cout << prefix << "Training on [" << imin << ", " << imax << "]." << endl;
  assert(imin <= imax);
  assert(eta0 > 0);
  for (int i=imin; i<=imax; i++)
    {
      double eta = eta0 / (1 + lambda * eta0 * t);
      trainOne(xp.at(i), yp.at(i), eta);
      t += 1;
    }
  /*
  cout << prefix << setprecision(6) << "wNorm=" << wnorm();
#if BIAS
  cout << " wBias=" << wBias;
#endif
  cout << endl;
  */
}

/// Perform a test pass
void 
SvmSgd::test(int imin, int imax, const xvec_t &xp, const yvec_t &yp, const char *prefix)
{
  cout << prefix << "Testing on [" << imin << ", " << imax << "]." << endl;
  assert(imin <= imax);
  double nerr = 0;
  double loss = 0;
  for (int i=imin; i<=imax; i++)
    testOne(xp.at(i), yp.at(i), &loss, &nerr);
  nerr = nerr / (imax - imin + 1);
  loss = loss / (imax - imin + 1);
  double cost = loss + 0.5 * lambda * wnorm();
  cout << prefix 
       << "Loss=" << setprecision(12) << loss
       << " Cost=" << setprecision(12) << cost 
       << " Misclassification=" << setprecision(4) << 100 * nerr << "%." 
       << endl;
}

/// Perform one epoch with fixed eta and return cost

double 
SvmSgd::evaluateEta(int imin, int imax, const xvec_t &xp, const yvec_t &yp, double eta)
{
  SvmSgd clone(*this); // take a copy of the current state
  assert(imin <= imax);
  for (int i=imin; i<=imax; i++)
    clone.trainOne(xp.at(i), yp.at(i), eta);
  double loss = 0;
  double cost = 0;
  for (int i=imin; i<=imax; i++)
    clone.testOne(xp.at(i), yp.at(i), &loss, 0);
  loss = loss / (imax - imin + 1);
  cost = loss + 0.5 * lambda * clone.wnorm();
  // cout << "Trying eta=" << eta << " yields cost " << cost << endl;
  return cost;
}

void 
SvmSgd::determineEta0(int imin, int imax, const xvec_t &xp, const yvec_t &yp)
{
  const double factor = 2.0;
  double loEta = 1;
  double loCost = evaluateEta(imin, imax, xp, yp, loEta);
  double hiEta = loEta * factor;
  double hiCost = evaluateEta(imin, imax, xp, yp, hiEta);
  if (loCost < hiCost)
    while (loCost < hiCost)
      {
        hiEta = loEta;
        hiCost = loCost;
        loEta = hiEta / factor;
        loCost = evaluateEta(imin, imax, xp, yp, loEta);
      }
  else if (hiCost < loCost)
    while (hiCost < loCost)
      {
        loEta = hiEta;
        loCost = hiCost;
        hiEta = loEta * factor;
        hiCost = evaluateEta(imin, imax, xp, yp, hiEta);
      }
  eta0 = loEta;
  //cout << "# Using eta0=" << eta0 << endl;
}

// interface functions
int svmsgdcount;
SvmSgd **svmsgd;

void InitializeSVM(int Dx, int L, double lambda)
{
	int i;
	svmsgdcount = L;
	svmsgd = new SvmSgd*[L];
	for (i = 0; i < L; i++)
		svmsgd[i] = new SvmSgd(Dx, lambda);
}

void TrainSVM(intype *X, biguint *Z, int Dx, int imin, int imax, int isvm, int stepcounter)
{
	static xvec_t xtrain;
	static yvec_t ytrain;

	xtrain.reserve(imax - imin + 1);
	ytrain.reserve(imax - imin + 1);

	static XVector x;

	while ((int)xtrain.size() < imax - imin + 1)
		xtrain.push_back(x);
	while ((int)ytrain.size() < imax - imin + 1)
		ytrain.push_back(0);

	/*
	char Xcach[(imax - imin + 1) * Dx];
	memcpy(Xcach, &(X[imin * Dx]), (imax - imin + 1) * Dx);

	biguint Zcach[imax - imin + 1];
	memcpy(Zcach, &(Z[imin]), (imax - imin + 1) * sizeof(biguint));
	*/

	intype *Xcach = &(X[imin * Dx]);
	biguint *Zcach = &(Z[imin]);

	int i, j;
	double y;

	for (i = 0; i <= imax - imin; i++)
	{
		for (j = 0; j < Dx; j++)
		{
			y = intype2double(Xcach[i * Dx + j]);
			x.set(j, y);
		}
		xtrain[i] = x;

		y = bui_getbit(&(Zcach[i]), isvm) ? 1 : -1;
	    if (y != +1 && y != -1)
	        assertfail("Label should be +1 or -1.");
		ytrain[i] = y;
	}

	//svmsgd[isvm]->t = stepcounter * (imax - imin);
	if (stepcounter == 0)
	{
		svmsgd[isvm]->t = 0;
		// determine eta0 using sample
		svmsgd[isvm]->determineEta0(0, min(1000, imax - imin), xtrain, ytrain);
	}

	svmsgd[isvm]->train(0, imax - imin, xtrain, ytrain);
}

double PredictSVM(int isvm, intype *x, int Dx)
{
	static XVector svecx;
	int i;
	for (i = 0; i < Dx; i++)
		svecx.set(i, intype2double(x[i]));

	double output;
	svmsgd[isvm]->testOne(svecx, 0/*ytrain[isvm].at(ix)*/, NULL, NULL, &output);
	return output > 0 ? 1 : 0;
}

void GetSVM(int isvm, double *buffer, int *index)
{
	int i, len = svmsgd[isvm]->w.size();
	for (i = 0; i < len; i++)
		buffer[*index + i] = svmsgd[isvm]->w.get(i);
	*index += len;

	buffer[*index + 0] = svmsgd[isvm]->lambda;
	buffer[*index + 1] = svmsgd[isvm]->eta0;
	buffer[*index + 2] = svmsgd[isvm]->wDivisor;
	buffer[*index + 3] = svmsgd[isvm]->wBias;
	buffer[*index + 4] = svmsgd[isvm]->t;
	*index += 5;
}

void SetSVM(int isvm, double *buffer, int *index)
{
	int i, len = svmsgd[isvm]->w.size();
	for (i = 0; i < len; i++)
		svmsgd[isvm]->w.set(i, buffer[*index + i]);
	*index += len;

	svmsgd[isvm]->lambda = buffer[*index + 0];
	svmsgd[isvm]->eta0 = buffer[*index + 1];
	svmsgd[isvm]->wDivisor = buffer[*index + 2];
	svmsgd[isvm]->wBias = buffer[*index + 3];
	svmsgd[isvm]->t = buffer[*index + 4];
	*index += 5;
}

