//============================================================================
// Name        : utilites.c
// Author      :
// Version     :
// Copyright   : Copyright (c) 2014 by Mehdi Alizadeh
// Description :
//============================================================================

#include "BA.h"

// used to synchronize all the parallel processes
// it uses a single MPI_Allreduce to force all the processes reach
// to this function at the same time.
int synchronizeprocesses()
{
	int mpirank;
	MPI_Comm_rank(MPI_COMM_WORLD, &mpirank);
	int mpisize;
	MPI_Comm_size(MPI_COMM_WORLD, &mpisize);

	int sum = 0;
	MPI_Allreduce(&mpirank, &sum, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
	if (sum == mpisize * (mpisize - 1) / 2)
		return 0;

	return -1;
}

// based on the total number of tasks, it groups the submodels
// in some taskgroups, we can gain some performance by sending
// small submodels in groups between the processes
int** dividetasks(int tasks, int processors, int *taskgroups)
{
	*taskgroups = tasks < processors ? tasks : processors;
	int **tasktable = (int**)malloc(*taskgroups * sizeof(int*) + 1);
	int itask, igroup;

	for (igroup = 0; igroup < *taskgroups; igroup++)
	{
		tasktable[igroup] = (int*)malloc(tasks * sizeof(int));
		for (itask = 0; itask < tasks + 1; itask++)
			tasktable[igroup][itask] = -1;
	}

	igroup = 0;
	for (itask = 0; itask < tasks; itask++)
	{
		// add task _itask to processor _iproc
		int i = 0;
		while (tasktable[igroup][i] != -1)
			i++;
		tasktable[igroup][i] = itask;

		igroup = (igroup + 1) % *taskgroups;
	}

	return tasktable;
}

// calculates the buffer size for each taskgroup by simply summing up
// the size of all the submodels in each taskgroup
int* calcbuffersizes(int Din, int Dout, int L, int taskgroups, int **tasks, int breakf)
{
	int svmsize = (Din + 5) * sizeof(double);
	int linearannsize = (L + 1) * sizeof(double) * (breakf ? 1 : Dout);

	int *buffersizes = (int*)malloc(taskgroups * sizeof(int));

	int igroup;
	for (igroup = 0; igroup < taskgroups; igroup++)
	{
		buffersizes[igroup] = 0;

		int itask = 0;
		while (tasks[igroup][itask] != -1)
		{
			if (tasks[igroup][itask] < L)
				buffersizes[igroup] += svmsize;
			else
				buffersizes[igroup] += linearannsize;

			itask++;
		}
	}

	return buffersizes;
}

void freadmatrix(char *filename, gsl_matrix *m)
{
	FILE* file = fopen(filename, "r");
	gsl_matrix_fread(file, m);
	fclose(file);
}

void freadbuffer(char *filename, unsigned char *buffer, long int len, long int offset)
{
	//printf("reading from %s size: %ld offset: %ld\n", filename, len, offset);

	FILE* file = fopen(filename, "r");
	fseek(file, offset, SEEK_SET);
	long int r = fread(buffer, 1, len, file);
	if (r != len)
		printf("unexpected read size. expected: %ld read: %ld\n", len, r);

	fclose(file);
}

void fwritebuffer(char *filename, unsigned char *buffer, int len)
{
	FILE* file = fopen(filename, "w");
	int r = fwrite(buffer, 1, len, file);
	if (r != len)
		printf("unexpected write size. expected: %d wrote: %d\n", len, r);
	fclose(file);
}

long int loaddatasets(configstruct *config, intype **Xin, outype **Xout, biguint **Z, int mpirank, int mpisize, int loadbinaries)
{
	long int trainN = config->Ntrain;
	int Din = config->Din;
	int Dout = config->Dout;
	int L = config->L;
	int loadL = config->initbinL;

	long int i, j;
	int p, k;
	long int loadcount = 0;

	i = 0;
	for (p = 0; p < mpisize; p++)
	{
		long int N = (trainN - i) / (mpisize - p);

		if (p == mpirank)
		{
			if (N > 0)
			{
				long int offset = 0;//98000000;

				*Xout = (outype*)malloc(N * Dout * sizeof(outype));
				freadbuffer(config->trainset, (unsigned char*)*Xout, N * Dout * sizeof(outype), (i + offset) * Dout * sizeof(outype));

				if (Din)
				{
					//*Xin = (intype*)malloc(N * Din * sizeof(intype));
					//freadbuffer(config->trainsetRBF, (unsigned char*)*Xin, N * Din * sizeof(intype), (i + offset) * Din * sizeof(intype));

					*Xin = (intype*)malloc(N * (Din + Dout) * sizeof(intype));

					if (config->trainsetRBF[0] == 'O' && config->trainsetRBF[1] == 'N' && config->trainsetRBF[2] == 'L' && config->trainsetRBF[3] == 'I')
					{
						int rbcount = Din;
						double sigma = config->rbfsigma;

						outype *R = (outype*)malloc(rbcount * Dout * sizeof(outype));
						freadbuffer(config->rbfcenters, (unsigned char*)R, rbcount * Dout * sizeof(outype), 0);
						calcRBF(*Xout, N, Dout, R, rbcount, sigma, *Xin);

						free(R);
					}
					else
					{
						intype *Xintemp = (intype*)malloc(N * Din * sizeof(intype));
						freadbuffer(config->trainsetRBF, (unsigned char*)Xintemp, N * Din * sizeof(intype), (i + offset) * Din * sizeof(intype));

						for (j = 0; j < N; j++)
						{
							memcpy(&((*Xin)[j * (Din + Dout)]), &(Xintemp[j * Din]), Din * sizeof(intype));
							for(k = 0; k < Dout; k++)
								(*Xin)[j * (Din + Dout) + Din + k] = 0;//double2intype(outype2double((*Xout)[j * Dout + k]));
						}

						free(Xintemp);
					}

					config->Din += config->Dout;
					Din += Dout;
				}
				else
				{
					config->Din = config->Dout;
					*Xin = (intype*)(*Xout);
					memcpy(&(config->trainsetRBF[0]), &(config->trainset[0]), 256);
				}

				*Z = (biguint*)calloc(N, sizeof(biguint));

				if (loadbinaries)
				{
					unsigned char *tempZ = (unsigned char*)malloc(N * loadL);
					freadbuffer(config->initbintrain, tempZ, N * loadL, (i + offset) * loadL);

					for (j = 0; j < N; j++)
						for (k = 0; k < L; k++)
							bui_setbit(&((*Z)[j]), k, tempZ[j * loadL + k]);
					free(tempZ);
				}
				else
				{
					srand(time(NULL));
					for (j = 0; j < N * L; j++)
						bui_setbit(&((*Z)[j / L]), j % L, rand() % 2);
				}
			}

			loadcount = N;
			printf("P%02d: loaded %ldx(%d+%d) datapoints and %ldx%d binary values.\n", mpirank, N, Din, Dout, N, L);
			break;
		}

		i += N;
	}

	return loadcount;
}

void print_matrix(char* name, gsl_matrix* m)
{
	fprintf(stdout, "%s:\n", name);

	int i, j;
	for (i = 0; i < 1/*m->size1*/; i++)
	{
		for (j = 0; j < m->size2; j++)
			fprintf(stdout, "%12.8f ", gsl_matrix_get(m, i, j));
		fprintf(stdout, "\n");
	}
}

// it is used mainly for debug purposes, and efficient way to
// compare to large buffers is to compare their hash code
// this method computes the fletcher32 hash code of a given buffer
unsigned int fletcher32(unsigned short int *data, int words)
{
	unsigned int sum1 = 0xffff, sum2 = 0xffff;

	while (words)
	{
		unsigned int tlen = words > 359 ? 359 : words;
		words -= tlen;
		do
		{
			sum2 += sum1 += *data++;
		} while (--tlen);

		sum1 = (sum1 & 0xffff) + (sum1 >> 16);
		sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	}

	/* Second reduction step to reduce sums to 16 bits */
	sum1 = (sum1 & 0xffff) + (sum1 >> 16);
	sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	return sum2 << 16 | sum1;
}

double* readarray(char *buf, int *len)
{
	int MAXLEN = 100;
	double* result = (double*)malloc(sizeof(double) * MAXLEN);

	int i, begin = 0, end, eob = 0;
	*len = 0;
	for (i = 0; i < MAXLEN && !eob; i++)
	{
		while (!((buf[begin] >= '0' && buf[begin] <= '9') || buf[begin] == '.'))
			begin++;

		end = begin;
		while ((buf[end] >= '0' && buf[end] <= '9') || buf[end] == '.')
			end++;

		if (buf[end] == '\0')
			eob = 1;

		buf[end] = '\0';
		char *part = &(buf[begin]);
		result[i] = atof(part);

		(*len)++;

		begin = end;
	}

	double* finalresult = (double*)malloc(sizeof(double) * (*len));
	memcpy(finalresult, result, sizeof(double) * (*len));

	return finalresult;
}

configstruct* loadconfig(char *path, int report)
{
	//*
	const int totallines = 23;
	char configbuf[totallines][256];
	int configidx = 0;

	FILE *ptr_file;
	char buf[1000];
	char *purebuf;
	ptr_file =fopen(path,"r");
	if (!ptr_file)
		return NULL;

	while (configidx < totallines && fgets(buf, 256, ptr_file) != NULL)
	{
		int len = strlen(buf);

		int i = 0;
		while (i < len)
			if (buf[i] == ' ' || buf[i] == '\t' || buf[i] == '\r' || buf[i] == '\n') i++;
			else break;

		int j = len - 1;
		while (j >= 0)
			if (buf[j] == ' ' || buf[j] == '\t' || buf[j] == '\r' || buf[j] == '\n') j--;
			else break;

		if (i <= j)
		{
			buf[j + 1] = '\0';
			purebuf = &(buf[i]);

			if (purebuf[0] != '#')
			{
				strcpy(configbuf[configidx], purebuf);
				configidx++;
			}
		}
	}

	fclose(ptr_file);

	// replace all $ with bit count
	for (configidx = 0; configidx < totallines; configidx++)
	{
		char* bitcount = configbuf[0];

		while (1)
		{
			int i = 0;
			int len = strlen(configbuf[configidx]);
			while (i < len && configbuf[configidx][i] != '$') i++;
			if (i == len) break;

			int j = 0;
			while (j < i)
			{
				buf[j] = configbuf[configidx][j]; j++;
			}

			int k = 0;
			while (bitcount[k] != '\0')
			{
				buf[j] = bitcount[k]; j++; k++;
			}

			i++;
			while (i <= len)
			{
				buf[j] = configbuf[configidx][i]; j++; i++;
			}

			strcpy(configbuf[configidx], buf);
		}
	}
	//*/

	configstruct *config = (configstruct*)malloc(sizeof(configstruct));

	//*
	int idx = 0;

	config->L = atoi(configbuf[idx++]);
	config->reporterrs = atoi(configbuf[idx++]);
	//config->maxit = atoi(configbuf[idx++]);

	config->epochs = atoi(configbuf[idx++]);
	config->minibatchcnt = atoi(configbuf[idx++]);
	config->mu = readarray(configbuf[idx++], &(config->maxit));

	config->N = atoi(configbuf[idx++]);
	config->Ntrain = atoi(configbuf[idx++]);
	config->Ntest = atoi(configbuf[idx++]);
	config->Din = atoi(configbuf[idx++]);
	config->Dout = atoi(configbuf[idx++]);

	config->K = atoi(configbuf[idx++]);
	config->k = atoi(configbuf[idx++]);

	config->initbinL = atoi(configbuf[idx++]);

	strcpy(config->trainset, configbuf[idx++]);
	strcpy(config->trainsetRBF, configbuf[idx++]);
	strcpy(config->testset, configbuf[idx++]);

	strcpy(config->neighbors, configbuf[idx++]);

	strcpy(config->initbintrain, configbuf[idx++]);
	strcpy(config->initbintest, configbuf[idx++]);

	strcpy(config->trainlabels, configbuf[idx++]);
	strcpy(config->testlabels, configbuf[idx++]);

	config->rbfsigma = atof(configbuf[idx++]);
	strcpy(config->rbfcenters, configbuf[idx++]);

	//*/

	if (report)
	{
		printf("%s:\n", path ? path : "hard-coded");
		//for (configidx = 0; configidx < totallines; configidx++)
		//	printf("%s%s", configbuf[configidx], configidx < 6 ? " " : "\n");

		printf("%d %d %d %d %ld %ld %ld %ld %d %d %d %d %d %f\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
		config->L,
		config->reporterrs,
		config->maxit,

		config->epochs,
		config->minibatchcnt,

		config->N,
		config->Ntrain,
		config->Ntest,
		config->Din,
		config->Dout,

		config->K,
		config->k,

		config->initbinL,

		config->rbfsigma,

		config->trainset,
		config->trainsetRBF,
		config->testset,

		config->neighbors,

		config->initbintrain,
		config->initbintest,

		config->trainlabels,
		config->testlabels,

		config->rbfcenters);

		for (idx = 0; idx < config->maxit; idx++)
			printf("%f ", config->mu[idx]);
		printf("\n");
	}

	return config;
}

int* loadperm(char *path, int i, int e, int p, long int m, int report)
{
	if (!path)
	{
		size_t size;
		FILE *stream;
		stream = open_memstream (&path, &size);

		fprintf (stream, "shuffle/shuffle-%d-%d-%d-%ld", i, e, p, m);

		fflush (stream);
		fclose (stream);
	}

	long int size = i * e * (p + m / p);
	int* permutation = (int*)malloc(size * sizeof(int));

	FILE *ptr_file;
	ptr_file =fopen(path,"r");

	int error = 0;

	if (!ptr_file)
	{
		error = 1;
		if (report)
			printf("could not open file %s. default permutation will be used.\n", path);
	}
	else
	{
		long int readsize = fread(permutation, sizeof(int), size, ptr_file);
		if (readsize != size)
		{
			error = 1;
			if (report)
				printf("invalid permutation file size in %s, expected %ld read %ld. default permutation will be used.\n", path, size, readsize);
		}
		fclose(ptr_file);
	}

	if (error)
	{
		long int idx = 0;
		int c;
		while (idx < size)
		{
			for (c = 0; c < p; c++)
				permutation[idx++] = c;
			for (c = 0; c < m / p; c++)
				permutation[idx++] = c;
		}
	}

	if (report)
	{
		int row = 0;
		long int c;
		printf("permutation row %d:", row);
		for (c = 0; c < p + m / p; c++)
			printf(" %02d", permutation[row * (p + m / p) + c]);
		printf("\n");
	}

	return permutation;
}

void calcranks(biguint *test, long int testN, biguint *query, long int queryN, biguint *nn, int *ranks, int worstcase)
{
	long int i, j;
	for (i = 0; i < queryN; i++)
	{
		ranks[i] = 0;
		int d = bui_fasthammingx64(&(query[i]), &(nn[i]));

		if (worstcase)
		{
			for (j = 0; j < testN; j++)
				if (bui_fasthammingx64(&(query[i]), &(test[j])) <= d)
					ranks[i]++;
		}
		else
		{
			if (d == 0)	continue;
			for (j = 0; j < testN; j++)
				if (bui_fasthammingx64(&(query[i]), &(test[j])) < d)
					ranks[i]++;
		}
	}
}

void calcgndbins(configstruct *config, unsigned int *gnd, biguint *gndbins, int loadbinaries)
{
	long int i, k;
	int L = config->L;
	int loadL = config->initbinL;
	int Din = config->Din;
	int Dout = config->Dout;
	long int N = config->Ntest;

	if (loadbinaries)
	{
		unsigned char *Z = (unsigned char*)malloc(N * L);
		for (i = 0; i < N; i++)
			freadbuffer(config->initbintrain, &(Z[i * L]), L, (long int)(gnd[i]) * loadL);
		for (i = 0; i < N * L; i++)
			bui_setbit(&(gndbins[i / L]), i % L, Z[i]);
		free(Z);
	}
	else
	{
		static intype *X = NULL;

		if (!X)
		{
			X = (intype*)malloc(N * Din * sizeof(intype));
			if (Din == Dout)
			{
				for (i = 0; i < N; i++)
					freadbuffer(config->trainset, (unsigned char*)(&(X[i * Din])), Din * sizeof(intype), (long int)(gnd[i]) * Din * sizeof(intype));
			}
			else if (config->trainsetRBF[0] == 'O' && config->trainsetRBF[1] == 'N' && config->trainsetRBF[2] == 'L' && config->trainsetRBF[3] == 'I')
			{
				outype *Xtemp = (outype*)malloc(N * Dout * sizeof(outype));
				for (i = 0; i < N; i++)
					freadbuffer(config->trainset, (unsigned char*)(&(Xtemp[i * Dout])), Dout * sizeof(intype), (long int)(gnd[i]) * Dout * sizeof(intype));

				int rbcount = Din - Dout;
				double sigma = config->rbfsigma;

				outype *R = (outype*)malloc(rbcount * Dout * sizeof(outype));
				freadbuffer(config->rbfcenters, (unsigned char*)R, rbcount * Dout * sizeof(outype), 0);
				calcRBF(Xtemp, N, Dout, R, rbcount, sigma, X);

				free(Xtemp);
				free(R);
			}
			else
			{
				outype *Xtemp = (outype*)malloc(Dout * sizeof(outype));
				for (i = 0; i < N; i++)
				{
					//freadbuffer(config->trainsetRBF, (unsigned char*)(&(X[i * Din])), (Din - Dout) * sizeof(intype), (long int)(gnd[i]) * (Din - Dout) * sizeof(intype));
					freadbuffer(config->trainsetRBF, (unsigned char*)(&(X[i * Din])), (Din - Dout) * sizeof(intype), (long int)(gnd[i]) * 4500 * sizeof(intype));
					freadbuffer(config->trainset, (unsigned char*)Xtemp, Dout * sizeof(outype), (long int)(gnd[i]) * Dout * sizeof(outype));

					for(k = 0; k < Dout; k++)
						X[i * Din + (Din - Dout) + k] = (intype)outype2double(Xtemp[k]);
				}
				free(Xtemp);
			}
		}

		HforwardSGD(X, gndbins, N, L, Din);
		//free(X);
	}
}

void calcRBF(outype *Xout, int N, int Dout, outype *R, int rbcount, double sigma, intype *Xin)
{
	int i, j, k;

	for (i = 0; i < N; i++)
	{
		for (j = 0; j < rbcount; j++)
		{
			double distance = 0, d;
			for (k = 0; k < Dout; k++)
			{
				d = ((double)(Xout[i * Dout + k]) - (double)(R[j * Dout + k])) / sigma;
				distance += d * d;
			}
			Xin[i * (rbcount + Dout) + j] = double2intype(exp(distance * -0.5));
		}

		for (k = 0; k < Dout; k++)
			Xin[i * (rbcount + Dout) + rbcount + k] = 0;//double2intype(outype2double(Xout[i * Dout + k]));
	}
}

int process_trained_models(int calculate_ranks, int calculate_AEerr, configstruct *config,
		int mpirank, double *modelbuffer, int modelsize, int totaltaskcount,
		intype *Xin, outype *Xout, biguint *Z, long int N, double *Nestederr, double *TOTALnestederr)
{
	int m;
	long int i;

	intype *Q = NULL;
	unsigned int *gnd = NULL;
	biguint *gndbins = NULL;
	biguint *QZ = NULL;

	int *localranksbest = NULL;
	int *totalranksbest = NULL;

	int *localrankswrst = NULL;
	int *totalrankswrst = NULL;

	if (calculate_ranks)
	{
		Q = (intype*)malloc(config->Ntest * config->Din * sizeof(intype));
		if (config->trainsetRBF[0] == 'O' && config->trainsetRBF[1] == 'N' && config->trainsetRBF[2] == 'L' && config->trainsetRBF[3] == 'I')
		{
			outype *Qtemp = (outype*)malloc(config->Ntest * config->Dout * sizeof(outype));
			freadbuffer(config->testset, (unsigned char*)Qtemp, config->Ntest * config->Dout * sizeof(outype), 0);

			int rbcount = config->Din - config->Dout;
			double sigma = config->rbfsigma;

			outype *R = (outype*)malloc(rbcount * config->Dout * sizeof(outype));
			freadbuffer(config->rbfcenters, (unsigned char*)R, rbcount * config->Dout * sizeof(outype), 0);
			calcRBF(Qtemp, config->Ntest, config->Dout, R, rbcount, sigma, Q);

			free(Qtemp);
			free(R);
		}
		else
			freadbuffer(config->testset, (unsigned char*)Q, config->Ntest * config->Din * sizeof(intype), 0);

		gnd = (unsigned int*)malloc(config->Ntest * config->K * sizeof(unsigned int));
		freadbuffer(config->neighbors, (unsigned char*)gnd, config->Ntest * config->K * sizeof(unsigned int), 0);

		// only keep the nearest neighbor
		unsigned int *tempgnd = gnd;
		gnd = (unsigned int*)malloc(config->Ntest * sizeof(unsigned int));

		for (i = 0; i < config->Ntest; i++)
			gnd[i] = tempgnd[i * config->K];

		free(tempgnd);
		gndbins = (biguint*)calloc(config->Ntest, sizeof(biguint));

		QZ = (biguint*)calloc(config->Ntest, sizeof(biguint));

		localranksbest = (int*)malloc((config->maxit + 1) * config->Ntest * sizeof(int));
		totalranksbest = (int*)malloc((config->maxit + 1) * config->Ntest * sizeof(int));

		localrankswrst = (int*)malloc((config->maxit + 1) * config->Ntest * sizeof(int));
		totalrankswrst = (int*)malloc((config->maxit + 1) * config->Ntest * sizeof(int));
	}

	for (m = 0; m <= config->maxit; m++)
	{
		if (m == 0)
		{
			if (!calculate_ranks)
				continue;

			unsigned char *tempQZ = (unsigned char*)malloc(config->Ntest * config->L);
			freadbuffer(config->initbintest, tempQZ, config->Ntest * config->L, 0);
			for (i = 0; i < config->Ntest * config->L; i++)
				bui_setbit(&(QZ[i / config->L]), i % config->L, tempQZ[i]);
			free(tempQZ);

			if (mpirank == ROOT_PROCESS)
				printf("loaded initial binaries\n");
		}
		else
		{
			char *modelpath = NULL;
			size_t size;
			FILE *stream;
			stream = open_memstream (&modelpath, &size);
			fprintf (stream, "model/model%02d", m);
			fflush (stream);
			fclose (stream);

			freadbuffer(modelpath, (unsigned char*)modelbuffer, modelsize, 0);

			int imodel;
			int imodelbuffindex = 0;
			for (imodel = 0; imodel < totaltaskcount; imodel++)
				if (imodel < config->L)
					setHmodel(imodel, modelbuffer, &imodelbuffindex);
				else
					setFmodel(imodel - config->L, modelbuffer, &imodelbuffindex);

			if (imodelbuffindex * sizeof(double) != modelsize)
				printf("inconsistent model %d %d\n", imodelbuffindex * (int)sizeof(double), modelsize);

			if (mpirank == ROOT_PROCESS)
				printf("loaded model%02d\n", m);

			HforwardSGD(Xin, Z, N, config->L, config->Din);
			if (mpirank == ROOT_PROCESS)
				printf("applied encoder%02d to baseset\n", m);

			if (calculate_ranks)
			{
				HforwardSGD(Q, QZ, config->Ntest, config->L, config->Din);
				if (mpirank == ROOT_PROCESS)
					printf("applied encoder%02d to queryset\n", m);
			}

			if (calculate_AEerr)
			{
				Ferror(Z, Xout, N, config->Dout, config->L, &(Nestederr[m - 1]));
				if (mpirank == ROOT_PROCESS)
					printf("applied decoder%02d and calculated local reconstruction error\n", m);
			}
		}

		if (calculate_ranks)
		{
			if (mpirank == ROOT_PROCESS)
				calcgndbins(config, gnd, gndbins, m == 0 ? 1 : 0);
			MPI_Bcast(gndbins, config->Ntest * sizeof(biguint), MPI_CHAR, ROOT_PROCESS, MPI_COMM_WORLD);

			calcranks(Z, N, QZ, config->Ntest, gndbins, &(localranksbest[m * config->Ntest]), 0);
			calcranks(Z, N, QZ, config->Ntest, gndbins, &(localrankswrst[m * config->Ntest]), 1);
			if (mpirank == ROOT_PROCESS)
				printf("updated local ranks\n");
		}
	}

	if (calculate_ranks)
	{
		MPI_Allreduce(&(localranksbest[0]), &(totalranksbest[0]), (config->maxit + 1) * config->Ntest, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
		MPI_Allreduce(&(localrankswrst[0]), &(totalrankswrst[0]), (config->maxit + 1) * config->Ntest, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

		if (mpirank == ROOT_PROCESS)
		{
			mkdir("rank/", S_IRWXO | S_IRWXG | S_IRWXU);

			fwritebuffer("rank/bestranks", (unsigned char*)(&(totalranksbest[0])), (config->maxit + 1) * config->Ntest * sizeof(int));
			fwritebuffer("rank/worstranks", (unsigned char*)(&(totalrankswrst[0])), (config->maxit + 1) * config->Ntest * sizeof(int));
			printf("updated total ranks and saved\n");
		}
	}

	if (calculate_AEerr)
	{
		MPI_Allreduce(&(Nestederr[0]), &(TOTALnestederr[0]), config->maxit, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
		if (mpirank == ROOT_PROCESS)
		{
			printf("total nested errors:\n");
			for (i = 0; i < config->maxit; i++)
				printf("%12.8f ", TOTALnestederr[i]);
			printf("\n");
		}
	}

	MPI_Finalize();
	return EXIT_SUCCESS;
}





